@@ -4,12 +4,15 @@
 
 import optparse
 import re
+from . import options as cmdoptions
 from . import command
 from . import utils
 from .msg import err_buf
 from . import scripts
 from . import completers as compl
 from . import bootstrap
+from . import corosync
+from .cibconfig import cib_factory
 
 
 class OptParser(optparse.OptionParser):
@@ -22,6 +25,9 @@ def _remove_completer(args):
         n = utils.list_cluster_nodes()
     except:
         n = []
+    for node in args[1:]:
+        if node in n:
+            n.remove(node)
     return scripts.param_completion_list('remove') + n
 
 
@@ -35,6 +41,17 @@ def script_args(args):
     return _nvpairs2parameters(args)
 
 
+def get_cluster_name():
+    cluster_name = None
+    if not bootstrap.service_is_active("corosync.service"):
+        name = corosync.get_values('totem.cluster_name')
+        if name:
+            cluster_name = name[0]
+    else:
+        cluster_name = cib_factory.get_property('cluster-name')
+    return cluster_name
+
+
 class Cluster(command.UI):
     '''
     Whole cluster management.
@@ -83,6 +100,40 @@ class Cluster(command.UI):
 
         # TODO: optionally stop services on all nodes or specific node
 
+    @command.skill_level('administrator')
+    def do_restart(self, context):
+        '''
+        Restarts the cluster services on this node
+        '''
+        self.do_stop(context)
+        self.do_start(context)
+
+    @command.skill_level('administrator')
+    def do_enable(self, context):
+        '''
+        Enable the cluster services on this node
+        '''
+        try:
+            utils.enable_service("pacemaker")
+            err_buf.info("Cluster services enabled")
+        except IOError as err:
+            context.fatal_error(str(err))
+
+        # TODO: optionally enable services on all nodes or specific node
+
+    @command.skill_level('administrator')
+    def do_disable(self, context):
+        '''
+        Disable the cluster services on this node
+        '''
+        try:
+            utils.disable_service("pacemaker")
+            err_buf.info("Cluster services disabled")
+        except IOError as err:
+            context.fatal_error(str(err))
+
+        # TODO: optionally disable services on all nodes or specific node
+
     def _args_implicit(self, context, args, name):
         '''
         handle early non-nvpair arguments as
@@ -133,42 +184,19 @@ Note:
     will be skipped.
 """, add_help_option=False)
 
-        parser.add_option("-h", "--help",action="store_true", dest="help", help="Show this help message")
-        parser.add_option("-q", "--quiet", action="store_true", dest="quiet",
-                          help="Be quiet (don't describe what's happening, just do it)")
-        parser.add_option("-y", "--yes", action="store_true", dest="yes_to_all",
-                          help='Answer "yes" to all prompts (use with caution, this is destructive, especially during the "storage" stage)')
-        parser.add_option("-t", "--template", dest="template",
-                          help='Optionally configure cluster with template "name" (currently only "ocfs2" is valid here)')
-        parser.add_option("-n", "--name", metavar="NAME", dest="name", default="hacluster",
-                          help='Set the name of the configured cluster.')
-        parser.add_option("-N", "--nodes", metavar="NODES", dest="nodes",
-                          help='Additional nodes to add to the created cluster. ' +
-                          'May include the current node, which will always be the initial cluster node.')
-        # parser.add_option("--quick-start", dest="quickstart", action="store_true", help="Perform basic system configuration (NTP, watchdog, /etc/hosts)")
-        parser.add_option("-S", "--enable-sbd", dest="diskless_sbd", action="store_true",
-                          help="Enable SBD even if no SBD device is configured (diskless mode)")
-        parser.add_option("-w", "--watchdog", dest="watchdog", metavar="WATCHDOG",
-                          help="Use the given watchdog device")
+        for _options in cmdoptions.CLUSTER_INIT_BASE_OPTIONS:
+            parser.add_option(*_options[0], **_options[-1])
 
         network_group = optparse.OptionGroup(parser, "Network configuration", "Options for configuring the network and messaging layer.")
-        network_group.add_option("-i", "--interface", dest="nic", metavar="IF",
-                                 help="Bind to IP address on interface IF")
-        network_group.add_option("-u", "--unicast", action="store_true", dest="unicast",
-                                 help="Configure corosync to communicate over unicast (UDP), and not multicast. " +
-                                 "Default is multicast unless an environment where multicast cannot be used is detected.")
-        network_group.add_option("-A", "--admin-ip", dest="admin_ip", metavar="IP",
-                                 help="Configure IP address as an administration virtual IP")
+        for _options in cmdoptions.CLUSTER_INIT_NET_OPTIONS:
+            network_group.add_option(*_options[0], **_options[-1])
         parser.add_option_group(network_group)
 
         storage_group = optparse.OptionGroup(parser, "Storage configuration", "Options for configuring shared storage.")
-        storage_group.add_option("-p", "--partition-device", dest="shared_device", metavar="DEVICE",
-                                 help='Partition this shared storage device (only used in "storage" stage)')
-        storage_group.add_option("-s", "--sbd-device", dest="sbd_device", metavar="DEVICE",
-                                 help="Block device to use for SBD fencing")
-        storage_group.add_option("-o", "--ocfs2-device", dest="ocfs2_device", metavar="DEVICE",
-                                 help='Block device to use for OCFS2 (only used in "vgfs" stage)')
+        for _options in cmdoptions.CLUSTER_INIT_STORAGE_OPTIONS:
+            storage_group.add_option(*_options[0], **_options[-1])
         parser.add_option_group(storage_group)
+
         try:
             options, args = parser.parse_args(list(args))
         except:
@@ -192,8 +220,17 @@ Note:
         #    parser.error("For a geo cluster, each cluster must have a unique name (use --name to set)")
         #    return False
 
+        qdevice = None
+        if options.qdevice:
+            qdevice = corosync.QDevice(
+                options.qdevice,
+                port=options.qdevice_port,
+                algo=options.qdevice_algo,
+                tie_breaker=options.qdevice_tie_breaker)
+
         bootstrap.bootstrap_init(
             cluster_name=options.name,
+            ui_context=context,
             nic=options.nic,
             ocfs2_device=options.ocfs2_device,
             shared_device=options.shared_device,
@@ -204,7 +241,10 @@ Note:
             admin_ip=options.admin_ip,
             yes_to_all=options.yes_to_all,
             unicast=options.unicast,
+            second_hb=options.second_hb,
+            ipv6=options.ipv6,
             watchdog=options.watchdog,
+            qdevice=qdevice,
             stage=stage,
             args=args)
 
@@ -216,7 +256,7 @@ Note:
             for node in nodelist:
                 if node == utils.this_node():
                     continue
-                bootstrap.status("Add node {} (may prompt for root password):".format(node))
+                bootstrap.status("\n\nAdd node {} (may prompt for root password):".format(node))
                 if not self._add_node(node, yes_to_all=options.yes_to_all):
                     return False
 
@@ -238,7 +278,7 @@ Stage can be one of:
 
 If stage is not specified, each stage will be invoked in sequence.
 """, add_help_option=False)
-        parser.add_option("-h", "--help",action="store_true", dest="help", help="Show this help message")
+        parser.add_option("-h", "--help", action="store_true", dest="help", help="Show this help message")
         parser.add_option("-q", "--quiet", help="Be quiet (don't describe what's happening, just do it)", action="store_true", dest="quiet")
         parser.add_option("-y", "--yes", help='Answer "yes" to all prompts (use with caution)', action="store_true", dest="yes_to_all")
         parser.add_option("-w", "--watchdog", dest="watchdog", metavar="WATCHDOG", help="Use the given watchdog device")
@@ -264,6 +304,7 @@ If stage is not specified, each stage will be invoked in sequence.
 
         bootstrap.bootstrap_join(
             cluster_node=options.cluster_node,
+            ui_context=context,
             nic=options.nic,
             quiet=options.quiet,
             yes_to_all=options.yes_to_all,
@@ -290,7 +331,7 @@ If stage is not specified, each stage will be invoked in sequence.
         Must be executed from a node in the existing cluster.
         '''
         parser = OptParser(usage="usage: add [options] [node ...]", add_help_option=False)
-        parser.add_option("-h", "--help",action="store_true", dest="help", help="Show this help message")
+        parser.add_option("-h", "--help", action="store_true", dest="help", help="Show this help message")
         parser.add_option("-y", "--yes", help='Answer "yes" to all prompts (use with caution)', action="store_true", dest="yes_to_all")
         try:
             options, args = parser.parse_args(list(args))
@@ -303,6 +344,7 @@ If stage is not specified, each stage will be invoked in sequence.
             if not self._add_node(node, yes_to_all=options.yes_to_all):
                 return False
 
+    @command.alias("delete")
     @command.completers_repeating(_remove_completer)
     @command.skill_level('administrator')
     def do_remove(self, context, *args):
@@ -310,7 +352,7 @@ If stage is not specified, each stage will be invoked in sequence.
         Remove the given node(s) from the cluster.
         '''
         parser = OptParser(usage="usage: remove [options] [<node> ...]", add_help_option=False)
-        parser.add_option("-h", "--help",action="store_true", dest="help", help="Show this help message")
+        parser.add_option("-h", "--help", action="store_true", dest="help", help="Show this help message")
         parser.add_option("-q", "--quiet", help="Be quiet (don't describe what's happening, just do it)", action="store_true", dest="quiet")
         parser.add_option("-y", "--yes", help='Answer "yes" to all prompts (use with caution)', action="store_true", dest="yes_to_all")
         parser.add_option("-c", "--cluster-node", dest="cluster_node", help="IP address or hostname of cluster node which will be deleted", metavar="HOST")
@@ -327,17 +369,46 @@ If stage is not specified, each stage will be invoked in sequence.
         if len(args) == 0:
             bootstrap.bootstrap_remove(
                 cluster_node=None,
+                ui_context=context,
                 quiet=options.quiet,
                 yes_to_all=options.yes_to_all)
         else:
             for node in args:
                 bootstrap.bootstrap_remove(
                     cluster_node=node,
+                    ui_context=context,
                     quiet=options.quiet,
                     yes_to_all=options.yes_to_all,
                     force=options.force)
         return True
 
+    @command.skill_level('administrator')
+    def do_rename(self, context, new_name):
+        '''
+        Rename the cluster.
+        '''
+        if not bootstrap.service_is_active("corosync.service"):
+            context.fatal_error("Can't rename cluster when cluster service is stopped")
+        old_name = cib_factory.get_property('cluster-name')
+        if old_name and new_name == old_name:
+            context.fatal_error("Expected a different name")
+
+        # Update config file with the new name on all nodes
+        nodes = utils.list_cluster_nodes()
+        corosync.set_value('totem.cluster_name', new_name)
+        if len(nodes) > 1:
+            nodes.remove(utils.this_node())
+            context.info("Copy cluster config file to \"{}\"".format(' '.join(nodes)))
+            corosync.push_configuration(nodes)
+
+        # Change the cluster-name property in the CIB
+        cib_factory.create_object("property", "cluster-name={}".format(new_name))
+        if not cib_factory.commit():
+            context.fatal_error("Change property cluster-name failed!")
+
+        # it's a safe way to give user a hints that need to restart service
+        context.info("To apply the change, restart the cluster service at convenient time")
+
     def _parse_clustermap(self, clusters):
         '''
         Helper function to parse the cluster map into a dictionary:
@@ -381,7 +452,7 @@ Cluster Description
   Name clusters using the --name parameter to
   crm bootstrap init.
 """, add_help_option=False)
-        parser.add_option("-h", "--help",action="store_true", dest="help", help="Show this help message")
+        parser.add_option("-h", "--help", action="store_true", dest="help", help="Show this help message")
         parser.add_option("-q", "--quiet", help="Be quiet (don't describe what's happening, just do it)", action="store_true", dest="quiet")
         parser.add_option("-y", "--yes", help='Answer "yes" to all prompts (use with caution)', action="store_true", dest="yes_to_all")
         parser.add_option("-a", "--arbitrator", help="IP address of geo cluster arbitrator", dest="arbitrator", metavar="IP")
@@ -410,7 +481,7 @@ Cluster Description
                 ticketlist = [t for t in re.split('[ ,;]+', options.tickets)]
             except ValueError:
                 parser.error("Invalid ticket list")
-        bootstrap.bootstrap_init_geo(options.quiet, options.yes_to_all, options.arbitrator, clustermap, ticketlist)
+        bootstrap.bootstrap_init_geo(options.quiet, options.yes_to_all, options.arbitrator, clustermap, ticketlist, ui_context=context)
         return True
 
     @command.name("geo_join")
@@ -421,7 +492,7 @@ Cluster Description
         Join this cluster to a geo configuration.
         '''
         parser = OptParser(usage="usage: geo-join [options]", add_help_option=False)
-        parser.add_option("-h", "--help",action="store_true", dest="help", help="Show this help message")
+        parser.add_option("-h", "--help", action="store_true", dest="help", help="Show this help message")
         parser.add_option("-q", "--quiet", help="Be quiet (don't describe what's happening, just do it)", action="store_true", dest="quiet")
         parser.add_option("-y", "--yes", help='Answer "yes" to all prompts (use with caution)', action="store_true", dest="yes_to_all")
         parser.add_option("-c", "--cluster-node", help="IP address of an already-configured geo cluster or arbitrator", dest="node", metavar="IP")
@@ -443,7 +514,7 @@ Cluster Description
         clustermap = self._parse_clustermap(options.clusters)
         if clustermap is None:
             parser.error("Invalid cluster description format")
-        bootstrap.bootstrap_join_geo(options.quiet, options.yes_to_all, options.node, clustermap)
+        bootstrap.bootstrap_join_geo(options.quiet, options.yes_to_all, options.node, clustermap, ui_context=context)
         return True
 
     @command.name("geo_init_arbitrator")
@@ -454,7 +525,7 @@ Cluster Description
         Make this node a geo arbitrator.
         '''
         parser = OptParser(usage="usage: geo-init-arbitrator [options]", add_help_option=False)
-        parser.add_option("-h", "--help",action="store_true", dest="help", help="Show this help message")
+        parser.add_option("-h", "--help", action="store_true", dest="help", help="Show this help message")
         parser.add_option("-q", "--quiet", help="Be quiet (don't describe what's happening, just do it)", action="store_true", dest="quiet")
         parser.add_option("-y", "--yes", help='Answer "yes" to all prompts (use with caution)', action="store_true", dest="yes_to_all")
         parser.add_option("-c", "--cluster-node", help="IP address of an already-configured geo cluster", dest="other", metavar="IP")
@@ -465,7 +536,7 @@ Cluster Description
         if options.help:
             parser.print_help()
             return
-        bootstrap.bootstrap_arbitrator(options.quiet, options.yes_to_all, options.other)
+        bootstrap.bootstrap_arbitrator(options.quiet, options.yes_to_all, options.other, ui_context=context)
         return True
 
     @command.completers_repeating(compl.call(scripts.param_completion_list, 'health'))
@@ -486,24 +557,26 @@ Cluster Description
         '''
         Quick cluster health status. Corosync status, DRBD status...
         '''
+
         stack = utils.cluster_stack()
         if not stack:
             err_buf.error("No supported cluster stack found (tried heartbeat|openais|corosync)")
         if utils.cluster_stack() == 'corosync':
-            print "Services:"
+            print("Name: {}\n".format(get_cluster_name()))
+            print("Services:")
             for svc in ["corosync", "pacemaker"]:
                 info = utils.service_info(svc)
                 if info:
-                    print "%-16s %s" % (svc, info)
+                    print("%-16s %s" % (svc, info))
                 else:
-                    print "%-16s unknown" % (svc)
+                    print("%-16s unknown" % (svc))
 
             rc, outp = utils.get_stdout(['corosync-cfgtool', '-s'], shell=False)
             if rc == 0:
-                print ""
-                print outp
+                print("")
+                print(outp)
             else:
-                print "Failed to get corosync status"
+                print("Failed to get corosync status")
 
     @command.completers_repeating(compl.choice(['10', '60', '600']))
     def do_wait_for_startup(self, context, timeout='10'):
@@ -520,9 +593,9 @@ Cluster Description
             context.fatal_error("Timed out waiting for cluster (rc = %s)" % (ret))
 
     @command.skill_level('expert')
-    def do_run(self, context, cmd):
+    def do_run(self, context, cmd, node=None):
         '''
-        Execute the given command on all nodes, report outcome
+        Execute the given command on all nodes/specific node, report outcome
         '''
         try:
             import parallax
@@ -534,15 +607,20 @@ Cluster Description
             context.fatal_error("python package parallax is needed for this command")
 
         hosts = utils.list_cluster_nodes()
+        if node and node in hosts:
+            hosts = [node]
         opts = parallax.Options()
-        for host, result in parallax.call(hosts, cmd, opts).iteritems():
+        for host, result in parallax.call(hosts, cmd, opts).items():
             if isinstance(result, parallax.Error):
                 err_buf.error("[%s]: %s" % (host, result))
             else:
                 if result[0] != 0:
-                    err_buf.error("[%s]: rc=%s\n%s\n%s" % (host, result[0], result[1], result[2]))
+                    err_buf.error("[%s]: rc=%s\n%s\n%s" % (host, result[0], utils.to_ascii(result[1]), utils.to_ascii(result[2])))
                 else:
-                    err_buf.ok("[%s]\n%s" % (host, result[1]))
+                    if not result[1]:
+                        err_buf.ok("[%s]" % host)
+                    else:
+                        err_buf.ok("[%s]\n%s" % (host, utils.to_ascii(result[1])))
 
     def do_copy(self, context, local_file, *nodes):
         '''
