@@ -39,6 +39,7 @@ def _tag_completions():
     "completer for tag: use in show"
     return ['tag:%s' % (t) for t in cib_factory.tag_list()]
 
+
 # Tab completion helpers
 _id_list = compl.call(cib_factory.id_list)
 _id_xml_list = compl.join(_id_list, compl.choice(['xml']))
@@ -54,9 +55,74 @@ _rsc_id_list = compl.call(cib_factory.rsc_id_list)
 _top_rsc_id_list = compl.call(cib_factory.top_rsc_id_list)
 _node_id_list = compl.call(cib_factory.node_id_list)
 _rsc_template_list = compl.call(cib_factory.rsc_template_list)
-_group_completer = compl.join(_f_prim_free_id_list, compl.choice(['params', 'meta']))
-_clone_completer = compl.choice(['params', 'meta'])
-_ms_completer = compl.choice(['params', 'meta'])
+_container_type = compl.choice(constants.container_type)
+
+
+def _group_completer(args):
+    '''
+    completer for group resource
+    '''
+    completing = args[-1]
+    id_list = cib_factory.f_prim_free_id_list()
+    if completing in id_list:
+        return [completing]
+    # complete resources id first
+    if len(args) == 3:
+        return [s for s in id_list if s not in args]
+    # complete meta or params attributes
+    key_words = ["meta", "params"]
+    keyw = last_keyword(args, key_words)
+    if keyw in key_words:
+        return _advanced_completer(args)
+
+    # otherwise, complete resources ids and some key words
+    return [s for s in id_list if s not in args] + _advanced_completer(args)
+
+
+def _advanced_completer(args):
+    '''
+    meta completers for group/ms/clone resource type 
+    '''
+    key_words = ["meta", "params"]
+    completing = args[-1]
+    resource_type = args[0]
+    return_list = []
+    if completing.endswith('='):
+        # TODO add some help messages
+        return []
+    keyw = last_keyword(args, key_words)
+    if keyw and keyw == "meta":
+        if resource_type == "group":
+            return_list = utils.filter_keys(constants.group_meta_attributes, args)
+        if resource_type == "clone":
+            return_list = utils.filter_keys(constants.clone_meta_attributes, args)
+        if resource_type in ["ms", "master"]:
+            return_list = utils.filter_keys(constants.ms_meta_attributes, args)
+    return return_list + key_words
+
+
+def _list_resource(args):
+    if len(args) > 3:
+        if args[2] == "remove":
+            res = cib_factory.f_prim_list_in_group(args[1])
+            if len(res) <= 1:
+                return []
+            else:
+                return res
+        if args[2] == "add":
+            return cib_factory.f_prim_free_id_list()
+
+
+def _list_resource_2(args):
+    if len(args) > 5:
+        return cib_factory.f_prim_list_in_group(args[1])
+
+
+def _pick_position(args):
+    if args[2] == "remove":
+        return []
+    else:
+        return ["after", "before"]
 
 
 def top_rsc_tmpl_id_list(args):
@@ -64,9 +130,8 @@ def top_rsc_tmpl_id_list(args):
 
 
 def ra_classes_or_tmpl(args):
-    if args[-1].startswith('@'):
-        return cib_factory.rsc_template_list()
-    return ui_ra.complete_class_provider_type(args)
+    return ui_ra.complete_class_provider_type(args) + \
+           ['@'+x for x in cib_factory.rsc_template_list()]
 
 
 def op_attr_list(args):
@@ -97,7 +162,8 @@ def _load_2nd_completer(args):
 def get_prim_token(words, n):
     for key in ("primitive", "rsc_template"):
         try:
-            return words[words.index(key) + n - 1]
+            if key in words:
+                return words[words.index(key) + n - 1]
         except IndexError:
             pass
     return ''
@@ -129,19 +195,22 @@ class CompletionHelp(object):
     lasttopic = ''
 
     @classmethod
-    def help(cls, topic, helptxt):
+    def help(cls, topic, helptxt, args):
         if cls.lasttopic == topic and \
                 time.time() - cls.laststamp < cls.timeout:
             return
         if helptxt:
             import readline
             cmdline = readline.get_line_buffer()
-            print "\n%s" % helptxt
-            if clidisplay.colors_enabled():
-                print "%s%s" % (term.render(clidisplay.prompt_noreadline(constants.prompt)),
-                                cmdline),
+            print("\n%s" % helptxt, end='')
+            if cmdline.split()[0] != args[0]:
+                prompt = '   > '
             else:
-                print "%s%s" % (constants.prompt, cmdline),
+                if clidisplay.colors_enabled():
+                    prompt = term.render(clidisplay.prompt_noreadline(constants.prompt))
+                else:
+                    prompt = constants.prompt
+            print("\n%s%s" % (prompt, cmdline), end=' ')
             cls.laststamp = time.time()
             cls.lasttopic = topic
 
@@ -153,11 +222,11 @@ def _prim_params_completer(agent, args):
     if completing.endswith('='):
         if len(completing) > 1 and options.interactive:
             topic = completing[:-1]
-            CompletionHelp.help(topic, agent.meta_parameter(topic))
+            CompletionHelp.help(topic, agent.meta_parameter(topic), args)
         return []
     elif '=' in completing:
         return []
-    return [s+'=' for s in agent.completion_params()]
+    return utils.filter_keys(agent.params(completion=True), args)
 
 
 def _prim_meta_completer(agent, args):
@@ -166,15 +235,48 @@ def _prim_meta_completer(agent, args):
         return ['meta']
     if '=' in completing:
         return []
-    return [s+'=' for s in constants.rsc_meta_attributes]
+    return utils.filter_keys(constants.rsc_meta_attributes, args)
 
 
 def _prim_op_completer(agent, args):
+    actions = agent.actions()
     completing = args[-1]
     if completing == 'op':
         return ['op']
     if args[-2] == 'op':
-        return list(constants.op_cli_names)
+        # append action items which in agent default actions
+        # monitor_Master will be mapped to "monitor role=Master"
+        # monitor_Slave will be mapped to "monitor role=Slave"
+        op_list = list(constants.op_cli_names)
+        if "monitor_Master" in actions:
+            op_list.append("monitor_Master")
+        if "monitor_Slave" in actions:
+            op_list.append("monitor_Slave")
+        # remove action items which not in default actions
+        for item in ["monitor", "demote", "promote", "notify"]:
+            if item not in actions:
+                op_list.remove(item)
+        # remove action items which already used
+        for item in op_list:
+            if item in args[:-2]:
+                op_list.remove(item)
+        return op_list
+    if args[-3] == 'op':
+        res = []
+        # list all of default items
+        if actions and actions[args[-2]]:
+            for k, v in list(actions[args[-2]].items()):
+                res += ["%s=%s" % (k, v)]
+            return res
+    args.pop()
+    # make sure all of default items can be completed
+    if args[-2] in actions:
+        res = []
+        for k, v in actions[args[-2]].items():
+            if args[-1].startswith(k+'='):
+                continue
+            res += ["%s=%s" % (k, v)]
+        return res
 
     return []
 
@@ -215,7 +317,7 @@ def primitive_complete_complex(args):
         "op": _prim_op_completer,
     }
 
-    keywords = completers_set.keys()
+    keywords = list(completers_set.keys())
     if len(args) == 4:  # <cmd> <id> <type> <?>
         return keywords
 
@@ -226,6 +328,149 @@ def primitive_complete_complex(args):
     return completers_set[last_keyw](agent, args) + keywords
 
 
+def container_helptxt(params, helptxt, topic):
+    for item in reversed(params):
+        if item in ["storage", "network", "docker", "rkt"]:
+            return helptxt[item][topic] + "\n"
+        if item == "port-mapping":
+            return helptxt["network"][item][topic] + "\n"
+
+
+def _container_remove_exist_keywords(args, _keywords):
+    for item in ["network", "primitive"]:
+        if item in args:
+            _keywords.remove(item)
+
+
+def _container_network_completer(args, _help, _keywords):
+    key_words = ["network", "port-mapping"]
+    completing = args[-1]
+    token = args[-2]
+    if completing.endswith("="):
+        return []
+    if completing in key_words:
+        return [completing]
+
+    tmp = list(_help["network"].keys())
+    # port-mapping is element, not a network option
+    tmp.remove("port-mapping")
+    network_keys = utils.filter_keys(tmp, args)
+    # bundle contain just one <network>/<primitive> element
+    _container_remove_exist_keywords(args, _keywords)
+
+    last_keyw = last_keyword(args, key_words)
+    if last_keyw == "network":
+        if token == "network":
+            return network_keys
+        else:
+            # complete port-mapping or other parts
+            return network_keys + ["port-mapping"] + _keywords
+
+    if last_keyw == "port-mapping":
+        mapping_required = ["id"]
+        mapping_params = args[utils.rindex(args, "port-mapping"):]
+        mapping_keys = utils.filter_keys(_help["network"]["port-mapping"].keys(), mapping_params)
+        if token == "port-mapping":
+            return mapping_keys
+        # required options must be completed
+        for s in mapping_required:
+            if utils.any_startswith(mapping_params, s+'=') is None:
+                return mapping_keys
+        # complete port-mapping or other parts
+        return mapping_keys + ["port-mapping"] + _keywords
+
+
+def _container_storage_completer(args, _help, _keywords):
+    completing = args[-1]
+    if completing.endswith("="):
+        return []
+    if completing == "storage":
+        return [completing]
+    if args[-2] == "storage":
+        return ["storage-mapping"]
+
+    storage_required = ["id", "target-dir"]
+    # get last storage part
+    mapping_params = args[utils.rindex(args, "storage-mapping"):]
+    storage_keys = utils.filter_keys(_help["storage"].keys(), mapping_params)
+
+    # required options must be completed
+    for s in storage_required:
+        if utils.any_startswith(mapping_params, s+"=") is None:
+            return storage_keys
+    # bundle contain just one <network>/<primitive> element
+    _container_remove_exist_keywords(args, _keywords)
+    # complete storage or other parts
+    return storage_keys + _keywords
+
+
+def _container_primitive_completer(args, _help, _keywords):
+    completing = args[-1]
+    if completing == "primitive":
+        return [completing]
+
+    _id_list = cib_factory.f_prim_free_id_list()
+    if _id_list is None:
+        return []
+    # bundle contain just one <network>/<primitive> element
+    _container_remove_exist_keywords(args, _keywords)
+    if args[-3] == "primitive" and args[-2] in _id_list:
+        return _keywords
+    return _id_list
+
+
+def _container_meta_completer(args, helptxt, _keywords):
+    completing = args[-1]
+    if completing.endswith("="):
+        return []
+    if completing == "meta":
+        return [completing]
+
+    # bundle contain just one <network>/<primitive> element
+    _container_remove_exist_keywords(args, _keywords)
+
+    return utils.filter_keys(constants.bundle_meta_attributes, args) + _keywords
+
+
+def container_complete_complex(args):
+    '''
+    Complete five parts:
+    container options, network, storage, primitive and meta
+    '''
+    container_options_required = ["image"]
+    completing = args[-1]
+    container_type = args[2]
+
+    completers_set = {
+        "network": _container_network_completer,
+        "storage": _container_storage_completer,
+        "primitive": _container_primitive_completer,
+        "meta": _container_meta_completer
+    }
+    keywords = list(completers_set.keys())
+    last_keyw = last_keyword(args, keywords)
+
+    # to show help messages
+    if completing.endswith('='):
+        if len(completing) > 1 and options.interactive:
+            topic = completing[:-1]
+            CompletionHelp.help(topic, container_helptxt(args, constants.container_helptxt, topic), args)
+        return []
+
+    container_options = utils.filter_keys(constants.container_helptxt[container_type].keys(), args)
+
+    # required options must be completed
+    for s in container_options_required:
+        if utils.any_startswith(args, s+'=') is None:
+            return container_options
+
+    if last_keyw is None:
+        return container_options + keywords
+
+    # to complete network, storage, primitive and meta
+    return completers_set[last_keyw](args, constants.container_helptxt, keywords)
+
+
 class CibConfig(command.UI):
     '''
     The configuration class
@@ -271,8 +516,8 @@ class CibConfig(command.UI):
     @command.name('_keywords')
     @command.skill_level('administrator')
     def do_keywords(self, context):
-        for k, v in sorted(constants.keywords.iteritems(), key=lambda v: v[0].lower()):
-            print "%-16s %s" % (k, v)
+        for k, v in sorted(iter(constants.keywords.items()), key=lambda v: v[0].lower()):
+            print("%-16s %s" % (k, v))
 
     @command.level(ui_ra.RA)
     def do_ra(self):
@@ -324,15 +569,15 @@ class CibConfig(command.UI):
 
         def print_value(v):
             if truth:
-                print utils.canonical_boolean(v)
+                print(utils.canonical_boolean(v))
             else:
-                print v
+                print(v)
         for p in properties:
             v = cib_factory.get_property_w_default(p)
             if v is not None:
                 print_value(v)
             elif truth:
-                print "false"
+                print("false")
             else:
                 context.fatal_error("%s: Property not set" % (p))
 
@@ -368,7 +613,7 @@ class CibConfig(command.UI):
 
     @command.skill_level('administrator')
     @command.completers(_f_group_id_list, compl.choice(['add', 'remove']),
-                        _prim_id_list, compl.choice(['after', 'before']), _prim_id_list)
+                        _list_resource, _pick_position, _list_resource_2)
     def do_modgroup(self, context, group_id, subcmd, prim_id, *args):
         """usage: modgroup <id> add <id> [after <id>|before <id>]
         modgroup <id> remove <id>"""
@@ -509,7 +754,7 @@ class CibConfig(command.UI):
         return set_obj.import_file(method, url)
 
     @command.skill_level('administrator')
-    @command.completers(compl.choice(gv_types.keys() + ['exportsettings']))
+    @command.completers(compl.choice(list(gv_types.keys()) + ['exportsettings']))
     def do_graph(self, context, *args):
         "usage: graph [<gtype> [<file> [<img_format>]]]"
         if args and args[0] == "exportsettings":
@@ -553,7 +798,7 @@ class CibConfig(command.UI):
         return cib_factory.default_timeouts(*args)
 
     @command.skill_level('administrator')
-    @command.completers(_id_list, _id_list)
+    @command.completers(_id_list)
     def do_rename(self, context, old_id, new_id):
         "usage: rename <old_id> <new_id>"
         return cib_factory.rename(old_id, new_id)
@@ -635,17 +880,24 @@ class CibConfig(command.UI):
     def do_schema(self, context, schema_st=None):
         "usage: schema [<schema>]"
         if not schema_st:
-            print cib_factory.get_schema()
+            print(cib_factory.get_schema())
             return True
         return cib_factory.change_schema(schema_st)
 
     def __conf_object(self, cmd, *args):
         "The configure object command."
-        if cmd in constants.cib_cli_map.values() and \
+        if cmd in list(constants.cib_cli_map.values()) and \
                 not cib_factory.is_elem_supported(cmd):
             common_err("%s not supported by the RNG schema" % cmd)
             return False
-        return cib_factory.create_object(cmd, *args)
+        if not args:
+            return cib_factory.create_object(cmd, *args)
+        if args[0].startswith("id="):
+            object_id = args[0][3:]
+        else:
+            object_id = args[0]
+        params = (object_id,) + args[1:]
+        return cib_factory.create_object(cmd, *params)
 
     @command.skill_level('administrator')
     @command.completers(_node_id_list, compl.choice(constants.node_attributes_keyw))
@@ -656,19 +908,36 @@ class CibConfig(command.UI):
         return self.__conf_object(context.get_command_name(), *args)
 
     @command.skill_level('administrator')
-    @command.completers_repeating(compl.null, ra_classes_or_tmpl, primitive_complete_complex)
+    @command.completers_repeating(compl.attr_id, ra_classes_or_tmpl, primitive_complete_complex)
     @command.alias('resource')
     def do_primitive(self, context, *args):
         """usage: primitive <rsc> {[<class>:[<provider>:]]<type>|@<template>}
-        [params <param>=<value> [<param>=<value>...]]
+        [[params] <param>=<value> [<param>=<value>...]]
         [meta <attribute>=<value> [<attribute>=<value>...]]
         [utilization <attribute>=<value> [<attribute>=<value>...]]
         [operations id_spec
-            [op op_type [<attribute>=<value>...] ...]]"""
+            [op op_type [<attribute>=<value>...]
+                        [[op_params] <param>=<value> [<param>=<value>...]]
+                        [op_meta <attribute>=<value> [<attribute>=<value>...]] ...]]"""
+        tmp = list(args)
+        for item in ['monitor_Master', 'monitor_Slave']:
+            if item in tmp:
+                idx = tmp.index(item)
+                tmp.remove(item)
+                tmp.insert(idx, "monitor")
+                tmp.insert(idx+1, "role=%s" % item.split('_')[1])
+        return self.__conf_object(context.get_command_name(), *tuple(tmp))
+
+    @command.completers_repeating(compl.attr_id, _container_type, container_complete_complex)
+    def do_bundle(self, context, *args):
+        """usage: bundle <bundle id> <container type> [<container option>...]
+        network [<network option>...]
+        storage [<storage option>...]
+        primitive <resource id> {[<class>:[<provider>:]]<type>|@<template>}"""
         return self.__conf_object(context.get_command_name(), *args)
 
     @command.skill_level('administrator')
-    @command.completers_repeating(compl.null, _group_completer)
+    @command.completers_repeating(compl.attr_id, _group_completer)
     def do_group(self, context, *args):
         """usage: group <name> <rsc> [<rsc>...]
         [params <param>=<value> [<param>=<value>...]]
@@ -676,7 +945,7 @@ class CibConfig(command.UI):
         return self.__conf_object(context.get_command_name(), *args)
 
     @command.skill_level('administrator')
-    @command.completers_repeating(compl.null, _f_children_id_list, _clone_completer)
+    @command.completers_repeating(compl.attr_id, _f_children_id_list, _advanced_completer)
     def do_clone(self, context, *args):
         """usage: clone <name> <rsc>
         [params <param>=<value> [<param>=<value>...]]
@@ -685,7 +954,7 @@ class CibConfig(command.UI):
 
     @command.alias('master')
     @command.skill_level('administrator')
-    @command.completers_repeating(compl.null, _f_children_id_list, _ms_completer)
+    @command.completers_repeating(compl.attr_id, _f_children_id_list, _advanced_completer)
     def do_ms(self, context, *args):
         """usage: ms <name> <rsc>
         [params <param>=<value> [<param>=<value>...]]
@@ -693,7 +962,7 @@ class CibConfig(command.UI):
         return self.__conf_object(context.get_command_name(), *args)
 
     @command.skill_level('administrator')
-    @command.completers_repeating(compl.null, ui_ra.complete_class_provider_type,
+    @command.completers_repeating(compl.attr_id, ui_ra.complete_class_provider_type,
                                   primitive_complete_complex)
     def do_rsc_template(self, context, *args):
         """usage: rsc_template <name> [<class>:[<provider>:]]<type>
@@ -705,7 +974,7 @@ class CibConfig(command.UI):
         return self.__conf_object(context.get_command_name(), *args)
 
     @command.skill_level('administrator')
-    @command.completers(compl.null, _top_rsc_id_list)
+    @command.completers(compl.attr_id, _top_rsc_id_list)
     def do_location(self, context, *args):
         """usage: location <id> <rsc> {node_pref|rules}
 
@@ -729,14 +998,14 @@ class CibConfig(command.UI):
 
     @command.alias('collocation')
     @command.skill_level('administrator')
-    @command.completers_repeating(compl.null, compl.null, top_rsc_tmpl_id_list)
+    @command.completers_repeating(compl.attr_id, compl.null, top_rsc_tmpl_id_list)
     def do_colocation(self, context, *args):
         """usage: colocation <id> <score>: <rsc>[:<role>] <rsc>[:<role>] ...
         [node-attribute=<node_attr>]"""
         return self.__conf_object(context.get_command_name(), *args)
 
     @command.skill_level('administrator')
-    @command.completers_repeating(compl.null,
+    @command.completers_repeating(compl.attr_id,
                                   compl.call(schema.rng_attr_values, 'rsc_order', 'kind'),
                                   top_rsc_tmpl_id_list)
     def do_order(self, context, *args):
@@ -745,7 +1014,7 @@ class CibConfig(command.UI):
         return self.__conf_object(context.get_command_name(), *args)
 
     @command.skill_level('administrator')
-    @command.completers_repeating(compl.null, compl.null, top_rsc_tmpl_id_list)
+    @command.completers_repeating(compl.attr_id, compl.null, top_rsc_tmpl_id_list)
     def do_rsc_ticket(self, context, *args):
         """usage: rsc_ticket <id> <ticket_id>: <rsc>[:<role>] [<rsc>[:<role>] ...]
         [loss-policy=<loss_policy_action>]"""
