@@ -14,6 +14,7 @@ import shlex
 import bz2
 import fnmatch
 import gc
+import ipaddress
 from contextlib import contextmanager
 from . import config
 from . import userdir
@@ -23,6 +24,42 @@ from . import term
 from .msg import common_warn, common_info, common_debug, common_err, err_buf
 
 
+mcast_regrex = r'2(?:2[4-9]|3\d)(?:\.(?:25[0-5]|2[0-4]\d|1\d\d|[1-9]\d?|0)){3}'
+
+
+def to_ascii(input_str):
+    """Convert the bytes string to a ASCII string
+    Usefull to remove accent (diacritics)"""
+    if input_str is None:
+        return input_str
+    if isinstance(input_str, str):
+        return input_str
+    try:
+        return str(input_str, 'utf-8')
+    except UnicodeDecodeError:
+        if config.core.debug or options.regression_tests:
+            import traceback
+            traceback.print_exc()
+        return input_str.decode('utf-8', errors='ignore')
+
+
+def filter_keys(key_list, args, sign="="):
+    """Return list item which not be completed yet"""
+    return [s+sign for s in key_list if any_startswith(args, s+sign) is None]
+
+
+def any_startswith(iterable, prefix):
+    """Return first element in iterable which startswith prefix, or None."""
+    for element in iterable:
+        if element.startswith(prefix):
+            return element
+    return None
+
+
+def rindex(iterable, value):
+    return len(iterable) - iterable[::-1].index(value) - 1
+
+
 def memoize(function):
     "Decorator to invoke a function once only for any argument"
     memoized = {}
@@ -73,17 +110,76 @@ def network_defaults(interface=None):
         if info[0] is None and len(sp) >= 5 and sp[0] == 'default' and sp[1] == 'via':
             info[0] = sp[4]
         if info[0] is not None and valfor(sp, 'dev') == info[0] and sp[0] != 'default':
+            src = valfor(sp, 'src')
             if sp[0].find('/') >= 0:
                 nw, length = sp[0].split('/')
-                info[1], info[2], info[3] = valfor(sp, 'src'), nw, length
+                info[1], info[2], info[3] = src, nw, length
+                break
             # we are reading /32 route entry
             else:
-                info[1], info[2], info[3] = valfor(sp, 'src'), sp[0], 32
+                info[1], info[2], info[3] = src, sp[0], 32
     if info[0] is None:
         raise ValueError("Failed to determine default network interface")
+    if info[1] is None:
+        ips = ip_in_local()
+        if len(ips) > 0:
+            info[1] = ips[0]
     return tuple(info)
 
 
+def network_all(with_mask=False):
+    """
+    returns all networks on local node
+    """
+    all_networks = []
+    _, outp = get_stdout("/sbin/ip -o route show")
+    for l in outp.split('\n'):
+        if re.search(r'\.0/[0-9]+ ', l):
+            if with_mask:
+                all_networks.append(l.split()[0])
+            else:
+                all_networks.append(l.split('/')[0])
+    return all_networks
+
+
+def network_v6_all():
+    _, outp = get_stdout("/sbin/ip -6 -o addr show")
+    dict_ = {}
+    for line in outp.split('\n'):
+        if re.search(r' ::1/| [Ff][Ee]80:', line):
+            # skip local address and link-local address
+            continue
+        dict_[line.split()[1]] = []
+    for line in outp.split('\n'):
+        if re.search(r' ::1/| [Ff][Ee]80:', line):
+            # skip local address and link-local address
+            continue
+        dict_[line.split()[1]].append(line.split()[3])
+    return dict_
+
+
+def ip_in_local(IPv6=False):
+    # short-circuit ip list handling for
+    # cloud providers in order to use
+    # metadata service instead
+    ips = iplist_for_cloud()
+    if len(ips) > 0:
+        return ips
+
+    if not IPv6:
+        regex = r' [0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}/[0-9]+ '
+    else:
+        regex = r' [0-9a-fA-F]{1,4}:.*:[0-9a-fA-F]{1,4}/[0-9]+ '
+
+    ip_local = []
+    _, outp = get_stdout("/sbin/ip addr show")
+    tmp = re.findall(regex, outp, re.M)
+    if tmp:
+        tmp = map(lambda x: x.split('/')[0].strip(), tmp)
+        ip_local += filter(lambda x: x != "127.0.0.1" and not re.search(r'^[Ff][Ee]80:', x), tmp)
+    return ip_local
+
+
 _cib_shadow = 'CIB_shadow'
 _cib_in_use = ''
 
@@ -205,7 +301,7 @@ def ask(msg):
 
     while True:
         try:
-            ans = raw_input(msg)
+            ans = input(msg)
         except EOFError:
             ans = 'n'
         if ans:
@@ -213,6 +309,7 @@ def ask(msg):
             if ans in 'yn':
                 return ans == 'y'
 
+
 # holds part of line before \ split
 # for a multi-line input
 _LINE_BUFFER = ''
@@ -232,12 +329,12 @@ def multi_input(prompt=''):
     _LINE_BUFFER = ''
     while True:
         try:
-            text = raw_input(prompt)
+            text = input(prompt)
         except EOFError:
             return None
         err_buf.incr_lineno()
         if options.regression_tests:
-            print ".INP:", text
+            print(".INP:", text)
             sys.stdout.flush()
             sys.stderr.flush()
         stripped = text.strip()
@@ -368,14 +465,17 @@ def pipe_string(cmd, s):
     cmd = add_sudo(cmd)
     common_debug("piping string to %s" % cmd)
     if options.regression_tests:
-        print ".EXT", cmd
+        print(".EXT", cmd)
     p = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE)
     try:
+        # communicate() expects encoded bytes
+        if isinstance(s, str):
+            s = s.encode('utf-8')
         p.communicate(s)
         p.wait()
         rc = p.returncode
     except IOError as msg:
-        if "Broken pipe" not in msg:
+        if "Broken pipe" not in str(msg):
             common_err(msg)
     return rc
 
@@ -390,39 +490,43 @@ def filter_string(cmd, s, stderr_on=True, shell=True):
     cmd = add_sudo(cmd)
     common_debug("pipe through %s" % cmd)
     if options.regression_tests:
-        print ".EXT", cmd
+        print(".EXT", cmd)
     p = subprocess.Popen(cmd,
                          shell=shell,
                          stdin=subprocess.PIPE,
                          stdout=subprocess.PIPE,
                          stderr=stderr)
     try:
+        # bytes expected here
+        if isinstance(s, str):
+            s = s.encode('utf-8')
         ret = p.communicate(s)
         if stderr_on == 'stdout':
-            outp = "\n".join(ret)
+            outp = b"\n".join(ret)
         else:
             outp = ret[0]
         p.wait()
         rc = p.returncode
-    except OSError, (errno, strerror):
-        if errno != os.errno.EPIPE:
-            common_err(strerror)
+    except OSError as err:
+        if err.errno != os.errno.EPIPE:
+            common_err(err.strerror)
         common_info("from: %s" % cmd)
-    except Exception, msg:
+    except Exception as msg:
         common_err(msg)
         common_info("from: %s" % cmd)
-    return rc, outp
+    return rc, to_ascii(outp)
 
 
-def str2tmp(s, suffix=".pcmk"):
+def str2tmp(_str, suffix=".pcmk"):
     '''
     Write the given string to a temporary file. Return the name
     of the file.
     '''
+    s = to_ascii(_str)
     fd, tmp = mkstemp(suffix=suffix)
     try:
         f = os.fdopen(fd, "w")
-    except IOError, msg:
+    except IOError as msg:
         common_err(msg)
         return
     f.write(s)
@@ -464,7 +568,7 @@ def create_tempfile(suffix='', dir=None):
 
 
 @contextmanager
-def open_atomic(filepath, mode="r", buffering=-1, fsync=False):
+def open_atomic(filepath, mode="r", buffering=-1, fsync=False, encoding=None):
     """ Open temporary file object that atomically moves to destination upon
     exiting.
 
@@ -483,7 +587,7 @@ def open_atomic(filepath, mode="r", buffering=-1, fsync=False):
     """
 
     with create_tempfile(dir=os.path.dirname(os.path.abspath(filepath))) as tmppath:
-        with open(tmppath, mode, buffering) as file:
+        with open(tmppath, mode, buffering, encoding=encoding) as file:
             try:
                 yield file
             finally:
@@ -498,9 +602,9 @@ def str2file(s, fname):
     Write a string to a file.
     '''
     try:
-        with open_atomic(fname, 'w') as dst:
-            dst.write(s)
-    except IOError, msg:
+        with open_atomic(fname, 'w', encoding='utf-8') as dst:
+            dst.write(to_ascii(s))
+    except IOError as msg:
         common_err(msg)
         return False
     return True
@@ -512,7 +616,7 @@ def file2str(fname, noerr=True):
     '''
     try:
         f = open(fname, "r")
-    except IOError, msg:
+    except IOError as msg:
         if not noerr:
             common_err(msg)
         return None
@@ -527,7 +631,7 @@ def file2list(fname):
     '''
     try:
         return open(fname).read().split('\n')
-    except IOError, msg:
+    except IOError as msg:
         common_err(msg)
         return None
 
@@ -541,7 +645,7 @@ def safe_open_w(fname):
                 return None
         try:
             f = open(fname, "w")
-        except IOError, msg:
+        except IOError as msg:
             common_err(msg)
             return None
     return f
@@ -578,7 +682,7 @@ def show_dot_graph(dotfile, keep_file=False, desc="transition graph"):
     if not keep_file:
         cmd = "(%s; rm -f %s)" % (cmd, dotfile)
     if options.regression_tests:
-        print ".EXT", cmd
+        print(".EXT", cmd)
     subprocess.Popen(cmd, shell=True, bufsize=0,
                      stdin=None, stdout=None, stderr=None, close_fds=True)
     common_info("starting %s to show %s" % (config.core.dotty, desc))
@@ -587,14 +691,14 @@ def show_dot_graph(dotfile, keep_file=False, desc="transition graph"):
 def ext_cmd(cmd, shell=True):
     cmd = add_sudo(cmd)
     if options.regression_tests:
-        print ".EXT", cmd
+        print(".EXT", cmd)
     common_debug("invoke: %s" % cmd)
     return subprocess.call(cmd, shell=shell)
 
 
 def ext_cmd_nosudo(cmd, shell=True):
     if options.regression_tests:
-        print ".EXT", cmd
+        print(".EXT", cmd)
     return subprocess.call(cmd, shell=shell)
 
 
@@ -632,7 +736,7 @@ def get_check_rc():
     return 2 which is the code for error. Otherwise, we
     pretend that errors are warnings.
     '''
-    return config.core.check_mode == "strict" and 2 or 1
+    return 2 if config.core.check_mode == "strict" else 1
 
 
 _LOCKDIR = ".lockdir"
@@ -650,12 +754,12 @@ def check_locker(lockdir):
         return
     try:
         os.kill(pid, 0)
-    except OSError, (errno, strerror):
-        if errno == os.errno.ESRCH:
+    except OSError as err:
+        if err.errno == os.errno.ESRCH:
             common_info("history: removing stale lock")
             rmdir_r(os.path.join(lockdir, _LOCKDIR))
         else:
-            common_err("%s: %s" % (_LOCKDIR, strerror))
+            common_err("%s: %s" % (_LOCKDIR, err.strerror))
 
 
 @contextmanager
@@ -672,9 +776,9 @@ def lock(lockdir):
                 os.makedirs(os.path.join(lockdir, _LOCKDIR))
                 str2file("%d" % os.getpid(), os.path.join(lockdir, _LOCKDIR, _PIDF))
                 return True
-            except OSError, (errno, strerror):
-                if errno != os.errno.EEXIST:
-                    common_err("Failed to acquire lock to %s: %s" % (lockdir, strerror))
+            except OSError as err:
+                if err.errno != os.errno.EEXIST:
+                    common_err("Failed to acquire lock to %s: %s" % (lockdir, err.strerror))
                     return False
                 time.sleep(0.1)
                 continue
@@ -689,7 +793,7 @@ def lock(lockdir):
             rmdir_r(os.path.join(lockdir, _LOCKDIR))
 
 
-def mkdirp(d, mode=0777):
+def mkdirp(d, mode=0o777):
     if os.path.isdir(d):
         return True
     os.makedirs(d, mode=mode)
@@ -697,7 +801,7 @@ def mkdirp(d, mode=0777):
 
 def pipe_cmd_nosudo(cmd):
     if options.regression_tests:
-        print ".EXT", cmd
+        print(".EXT", cmd)
     proc = subprocess.Popen(cmd,
                             shell=True,
                             stdout=subprocess.PIPE,
@@ -706,8 +810,8 @@ def pipe_cmd_nosudo(cmd):
     proc.wait()
     rc = proc.returncode
     if rc != 0:
-        print outp
-        print err_outp
+        print(outp)
+        print(err_outp)
     return rc
 
 
@@ -722,14 +826,14 @@ def get_stdout(cmd, input_s=None, stderr_on=True, shell=True):
     else:
         stderr = subprocess.PIPE
     if options.regression_tests:
-        print ".EXT", cmd
+        print(".EXT", cmd)
     proc = subprocess.Popen(cmd,
                             shell=shell,
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=stderr)
     stdout_data, stderr_data = proc.communicate(input_s)
-    return proc.returncode, stdout_data.strip()
+    return proc.returncode, to_ascii(stdout_data).strip()
 
 
 def get_stdout_stderr(cmd, input_s=None, shell=True):
@@ -737,14 +841,14 @@ def get_stdout_stderr(cmd, input_s=None, shell=True):
     Run a cmd, return (rc, stdout, stderr)
     '''
     if options.regression_tests:
-        print ".EXT", cmd
+        print(".EXT", cmd)
     proc = subprocess.Popen(cmd,
                             shell=shell,
                             stdin=input_s and subprocess.PIPE or None,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
     stdout_data, stderr_data = proc.communicate(input_s)
-    return proc.returncode, stdout_data.strip(), stderr_data.strip()
+    return proc.returncode, to_ascii(stdout_data).strip(), to_ascii(stderr_data).strip()
 
 
 def stdout2list(cmd, stderr_on=True, shell=True):
@@ -755,8 +859,7 @@ def stdout2list(cmd, stderr_on=True, shell=True):
     rc, s = get_stdout(add_sudo(cmd), stderr_on=stderr_on, shell=shell)
     if not s:
         return rc, []
-    else:
-        return rc, s.split('\n')
+    return rc, s.split('\n')
 
 
 def append_file(dest, src):
@@ -764,7 +867,7 @@ def append_file(dest, src):
     try:
         open(dest, "a").write(open(src).read())
         return True
-    except IOError, msg:
+    except IOError as msg:
         common_err("append %s to %s: %s" % (src, dest, msg))
         return False
 
@@ -812,7 +915,7 @@ def wait4dc(what="", show_progress=True):
         delaymsec = crm_msec(delay)
         if delaymsec > 0:
             common_info("The crmd-transition-delay is configured. Waiting %d msec before check DC status." % delaymsec)
-            time.sleep(delaymsec / 1000)
+            time.sleep(delaymsec // 1000)
     cnt = 0
     output_started = 0
     init_sleep = 0.25
@@ -950,7 +1053,7 @@ def crm_msec(t):
         mult, div = convtab[q]
     except KeyError:
         return -1
-    return (int(r.group(1))*mult)/div
+    return (int(r.group(1))*mult) // div
 
 
 def crm_time_cmp(a, b):
@@ -999,8 +1102,8 @@ def convert2ints(l):
     try:
         if isinstance(l, (tuple, list)):
             return [int(x) for x in l]
-        else:  # it's a string then
-            return int(l)
+        # it's a string then
+        return int(l)
     except ValueError:
         return None
 
@@ -1027,7 +1130,7 @@ def is_process(s):
     for pid in pids:
         try:
             cmdline = open(join('/proc', pid, 'cmdline'), 'rb').read()
-            procname = basename(cmdline.replace('\x00', ' ').split(' ')[0])
+            procname = basename(to_ascii(cmdline).replace('\x00', ' ').split(' ')[0])
             if procname == s:
                 return True
         except EnvironmentError:
@@ -1089,7 +1192,7 @@ def edit_file_ext(fname, template=''):
             f2.close()
         finally:
             os.unlink(tmpfile)
-    except OSError, e:
+    except OSError as e:
         raise IOError(e)
 
 
@@ -1099,7 +1202,7 @@ def need_pager(s, w, h):
     for l in s.split('\n'):
         # need to remove color codes
         l = re.sub(r'\${\w+}', '', l)
-        cnt += int(ceil((len(l) + 0.5)/w))
+        cnt += int(ceil((len(l) + 0.5) / w))
         if cnt >= h:
             return True
     return False
@@ -1126,13 +1229,15 @@ def page_string(s):
     'Page string rendered for TERM.'
     if not s:
         return
+    constants.need_reset = True
     w, h = get_winsize()
     if not need_pager(s, w, h):
-        print term_render(s)
+        print(term_render(s))
     elif not config.core.pager or not can_ask() or options.batch:
-        print term_render(s)
+        print(term_render(s))
     else:
-        pipe_string(get_pager_cmd(), term_render(s))
+        pipe_string(get_pager_cmd(), term_render(s).encode('utf-8'))
+    constants.need_reset = False
 
 
 def page_gen(g):
@@ -1179,11 +1284,11 @@ def multicolumn(l):
     for s in l:
         if len(s) > max_len:
             max_len = len(s)
-    cols = w/(max_len + min_gap)  # approx.
+    cols = w // (max_len + min_gap)  # approx.
     if not cols:
         cols = 1
-    col_len = w/cols
-    for i in range(len(l)/cols + 1):
+    col_len = w // cols
+    for i in range(len(l) // cols + 1):
         s = ''
         for j in range(i * cols, (i + 1) * cols):
             if not j < len(l):
@@ -1195,7 +1300,7 @@ def multicolumn(l):
             else:
                 s = "%s%-*s" % (s, col_len, l[j])
         if s:
-            print s
+            print(s)
 
 
 def find_value(pl, name):
@@ -1265,8 +1370,7 @@ def total_seconds(td):
     """
     if hasattr(datetime.timedelta, 'total_seconds'):
         return td.total_seconds()
-    else:
-        return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) / 10**6
+    return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 10**6) // 10**6
 
 
 def datetime_to_timestamp(dt):
@@ -1312,14 +1416,14 @@ def parse_time(t):
         else:
             # convert to UTC from local time
             dt = dt - tz.tzlocal().utcoffset(dt)
-    except ValueError, msg:
+    except ValueError as msg:
         common_err("parse_time %s: %s" % (t, msg))
         return None
-    except ImportError, msg:
+    except ImportError as msg:
         try:
             tm = time.strptime(t)
             dt = datetime.datetime(*tm[0:7])
-        except ValueError, msg:
+        except ValueError as msg:
             common_err("no dateutil, please provide times as printed by date(1)")
             return None
     return dt
@@ -1338,18 +1442,17 @@ def parse_to_timestamp(t):
 
         if datetime_is_aware(dt):
             return datetime_to_timestamp(dt)
-        else:
-            # convert to UTC from local time
-            return total_seconds(dt - tz.tzlocal().utcoffset(dt) - datetime.datetime(1970, 1, 1))
-    except ValueError, msg:
+        # convert to UTC from local time
+        return total_seconds(dt - tz.tzlocal().utcoffset(dt) - datetime.datetime(1970, 1, 1))
+    except ValueError as msg:
         common_err("parse_time %s: %s" % (t, msg))
         return None
-    except ImportError, msg:
+    except ImportError as msg:
         try:
             tm = time.strptime(t)
             dt = datetime.datetime(*tm[0:7])
             return datetime_to_timestamp(dt)
-        except ValueError, msg:
+        except ValueError as msg:
             common_err("no dateutil, please provide times as printed by date(1)")
             return None
 
@@ -1363,18 +1466,18 @@ def save_graphviz_file(ini_f, attr_d):
         return False
     try:
         f = open(ini_f, "wb")
-    except IOError, msg:
+    except IOError as msg:
         common_err(msg)
         return False
-    import ConfigParser
-    p = ConfigParser.SafeConfigParser()
-    for section, sect_d in attr_d.iteritems():
+    import configparser
+    p = configparser.SafeConfigParser()
+    for section, sect_d in attr_d.items():
         p.add_section(section)
-        for n, v in sect_d.iteritems():
+        for n, v in sect_d.items():
             p.set(section, n, v)
     try:
         p.write(f)
-    except IOError, msg:
+    except IOError as msg:
         common_err(msg)
         return False
     f.close()
@@ -1388,11 +1491,11 @@ def load_graphviz_file(ini_f):
     '''
     if not os.path.isfile(ini_f):
         return True, None
-    import ConfigParser
-    p = ConfigParser.SafeConfigParser()
+    import configparser
+    p = configparser.SafeConfigParser()
     try:
         p.read(ini_f)
-    except Exception, msg:
+    except Exception as msg:
         common_err(msg)
         return False, None
     _graph_d = {}
@@ -1424,7 +1527,7 @@ def get_pcmk_version(dflt):
             else:
                 version = s.split()[2]
             common_debug("found pacemaker version: %s" % version)
-    except Exception, msg:
+    except Exception as msg:
         common_warn("could not get the pacemaker version, bad installation?")
         common_warn(msg)
     return version
@@ -1442,7 +1545,7 @@ def get_cib_property(cib_f, attr, dflt):
     ver = dflt  # return some version in any case
     try:
         f = open(cib_f, "r")
-    except IOError, msg:
+    except IOError as msg:
         common_err(msg)
         return ver
     state = 0
@@ -1471,14 +1574,15 @@ def get_cib_attributes(cib_f, tag, attr_l, dflt_l):
     val_patt_l = [re.compile('%s="([^"]+)"' % x) for x in attr_l]
     val_l = []
     try:
-        f = open(cib_f).read()
-    except IOError, msg:
+        f = open(cib_f, "rb").read()
+    except IOError as msg:
         common_err(msg)
         return dflt_l
     if os.path.splitext(cib_f)[-1] == '.bz2':
-        cib_s = bz2.decompress(f)
+        cib_bits = bz2.decompress(f)
     else:
-        cib_s = f
+        cib_bits = f
+    cib_s = to_ascii(cib_bits)
     for s in cib_s.split('\n'):
         if s.startswith(open_t):
             i = 0
@@ -1566,7 +1670,7 @@ def fetch_opts(args, opt_l):
         re_opt = re.compile("^%s$" % opt_l[0][1:])
         del opt_l[0]
     l = []
-    for i in reversed(range(len(args))):
+    for i in reversed(list(range(len(args)))):
         if (args[i] in opt_l) or (re_opt and re_opt.search(args[i])):
             l.append(args.pop())
         else:
@@ -1650,7 +1754,6 @@ def list_cluster_nodes():
     '''
     Returns a list of nodes in the cluster.
     '''
-
     def getname(toks):
         if toks and len(toks) >= 2:
             return toks[1]
@@ -1658,10 +1761,23 @@ def list_cluster_nodes():
 
     try:
         rc, outp = stdout2list(['crm_node', '-l'], stderr_on=False, shell=False)
-        if rc != 0:
-            raise ValueError("Error listing cluster nodes: crm_node (rc=%d)" % (rc))
-        return [x for x in [getname(line.split()) for line in outp] if x and x != '(null)']
-    except OSError, msg:
+        if rc == 0:
+            return [x for x in [getname(line.split()) for line in outp] if x and x != '(null)']
+
+        from . import xmlutil
+        node_list = []
+        cib = xmlutil.cibdump2elem()
+        if cib is None:
+            return None
+        for node in cib.xpath('/cib/configuration/nodes/node'):
+            name = node.get('uname') or node.get('id')
+            if node.get('type') == 'remote':
+                srv = cib.xpath("//primitive[@id='%s']/instance_attributes/nvpair[@name='server']" % (name))
+                if srv:
+                    continue
+            node_list.append(name)
+        return node_list
+    except OSError as msg:
         raise ValueError("Error listing cluster nodes: %s" % (msg))
 
 
@@ -1754,7 +1870,7 @@ def unquote(s):
     becomes 'a b c'
     """
     sp = shlex.split(s)
-    if len(sp) > 0:
+    if sp:
         return sp[0]
     return ""
 
@@ -1764,14 +1880,15 @@ def parse_sysconfig(sysconfig_file):
     Reads a sysconfig file into a dict
     """
     ret = {}
-    vre = re.compile(r"(\S+)\s*=\s*(.*)")
     if os.path.isfile(sysconfig_file):
         for line in open(sysconfig_file).readlines():
             if line.lstrip().startswith('#'):
                 continue
-            m = vre.match(line)
-            if m:
-                ret[m.group(1)] = unquote(m.group(2))
+            try:
+                key, val = line.split("=", 1)
+                ret[key] = unquote(val)
+            except ValueError:
+                pass
     return ret
 
 
@@ -1780,7 +1897,6 @@ def sysconfig_set(sysconfig_file, **values):
     Set the values in the sysconfig file, updating the variables
     if they exist already, appending them if not.
     """
-    vre = re.compile(r"(\S+)\s*=\s*(.*)")
     outp = ""
     if os.path.isfile(sysconfig_file):
         for line in open(sysconfig_file).readlines():
@@ -1788,17 +1904,20 @@ def sysconfig_set(sysconfig_file, **values):
                 outp += line
             else:
                 matched = False
-                m = vre.match(line)
-                if m:
-                    for k, v in values.iteritems():
-                        if k == m.group(1):
-                            matched = True
-                            outp += '%s=%s\n' % (k, doublequote(v))
-                            del values[k]
-                            break
+            try:
+                key, _ = line.split("=", 1)
+                for k, v in values.items():
+                    if k == key:
+                        matched = True
+                        outp += '%s=%s\n' % (k, doublequote(v))
+                        del values[k]
+                        break
                 if not matched:
                     outp += line
-    for k, v in values.iteritems():
+            except ValueError:
+                outp += line
+
+    for k, v in values.items():
         outp += '%s=%s\n' % (k, doublequote(v))
     str2file(outp, sysconfig_file)
 
@@ -1814,7 +1933,7 @@ def remote_diff_slurp(nodes, filename):
     opts = parallax.Options()
     opts.localdir = tmpdir
     dst = os.path.basename(filename)
-    return parallax.slurp(nodes, filename, dst, opts).items()
+    return list(parallax.slurp(nodes, filename, dst, opts).items())
 
 
 def remote_diff_this(local_path, nodes, this_node):
@@ -1862,12 +1981,12 @@ def remote_checksum(local_path, nodes, this_node):
         if isinstance(result, parallax.Error):
             raise ValueError(str(result))
 
-    print "%-16s  SHA1 checksum of %s" % ('Host', local_path)
+    print("%-16s  SHA1 checksum of %s" % ('Host', local_path))
     if this_node not in nodes:
-        print "%-16s: %s" % (this_node, hashlib.sha1(open(local_path).read()).hexdigest())
+        print("%-16s: %s" % (this_node, hashlib.sha1(open(local_path).read()).hexdigest()))
     for host, result in by_host:
         _, _, _, path = result
-        print "%-16s: %s" % (host, hashlib.sha1(open(path).read()).hexdigest())
+        print("%-16s: %s" % (host, hashlib.sha1(open(path).read()).hexdigest()))
 
 
 def cluster_copy_file(local_path, nodes=None):
@@ -1887,7 +2006,7 @@ def cluster_copy_file(local_path, nodes=None):
     ok = True
     for host, result in parallax.copy(nodes,
                                       local_path,
-                                      local_path, opts).iteritems():
+                                      local_path, opts).items():
         if isinstance(result, parallax.Error):
             err_buf.error("Failed to push %s to %s: %s" % (local_path, host, result))
             ok = False
@@ -1895,6 +2014,7 @@ def cluster_copy_file(local_path, nodes=None):
             err_buf.ok(host)
     return ok
 
+
 # a set of fnmatch patterns to match attributes whose values
 # should be obscured as a sequence of **** when printed
 _obscured_nvpairs = []
@@ -1919,4 +2039,131 @@ def obscure(obscure_list):
         _obscured_nvpairs = prev
 
 
+def valid_ip_addr(addr, version=4):
+    import socket
+    try:
+        if version == 4:
+            socket.inet_pton(socket.AF_INET, addr)
+        elif version == 6:
+            socket.inet_pton(socket.AF_INET6, addr)
+        else:
+            return False
+    except socket.error:
+        return False
+    return True
+
+
+def get_ipv6_network(addr_with_mask):
+    return str(ipaddress.ip_interface(addr_with_mask).network.network_address)
+
+
+def gen_nodeid_from_ipv6(addr):
+    return int(ipaddress.ip_address(addr)) % 1000000000
+
+
+def ip_in_network(addr, net):
+    return ipaddress.ip_address(addr) in ipaddress.ip_interface(net).network
+
+
+class IP:
+    def __init__(self, addr):
+        self.addr = ipaddress.ip_address(addr)
+
+    def version(self):
+        return self.addr.version
+
+# Set by detect_cloud() or iplist_for_cloud()
+# to avoid multiple requests
+_ip_for_cloud = None
+
+
+def _cloud_metadata_request(uri, headers={}):
+    try:
+        import urllib2 as urllib
+    except ImportError:
+        import urllib.request as urllib
+    req = urllib.Request(uri)
+    for header, value in headers.items():
+        req.add_header(header, value)
+    try:
+        resp = urllib.urlopen(req, timeout=5)
+        content = resp.read()
+        if type(content) != str:
+            return content.decode('utf-8').strip()
+        return content.strip()
+    except urllib.URLError:
+        return None
+
+
+@memoize
+def detect_cloud():
+    """
+    Tries to determine which (if any) cloud environment
+    the cluster is running on.
+
+    This is mainly done using dmidecode.
+
+    If the host cannot be determined, this function
+    returns None. Otherwise, it returns a string
+    identifying the platform.
+
+    These are the currently known platforms:
+
+    * amazon-web-services
+    * microsoft-azure
+    * google-cloud-platform
+
+    """
+    global _ip_for_cloud
+
+    if not is_program("dmidecode"):
+        return None
+    rc, system_version = get_stdout("dmidecode -s system-version")
+    if re.search(r"\<.*\.amazon\>", system_version) is not None:
+        return "amazon-web-services"
+    if rc != 0:
+        return None
+    rc, system_manufacturer = get_stdout("dmidecode -s system-manufacturer")
+    if rc == 0 and "microsoft corporation" in system_manufacturer.lower():
+        # To detect azure we also need to make an API request
+        result = _cloud_metadata_request(
+            "http://169.254.169.254/metadata/instance/network/interface/0/ipv4/ipAddress/0/privateIpAddress?api-version=2017-08-01&format=text",
+            headers={"Metadata": "true"})
+        if result:
+            _ip_for_cloud = result
+            return "microsoft-azure"
+    rc, bios_vendor = get_stdout("dmidecode -s bios-vendor")
+    if rc == 0 and "Google" in bios_vendor:
+        # To detect GCP we also need to make an API request
+        result = _cloud_metadata_request(
+            "http://metadata.google.internal/computeMetadata/v1/instance/network-interfaces/0/ip",
+            headers={"Metadata-Flavor": "Google"})
+        if result:
+            _ip_for_cloud = result
+            return "google-cloud-platform"
+    return None
+
+
+def iplist_for_cloud():
+    """
+    Return list of viable IPs for the
+    current cloud provider (if any)
+    """
+    global _ip_for_cloud
+    provider = detect_cloud()
+    if _ip_for_cloud is not None:
+        return [_ip_for_cloud]
+    if provider == "amazon-web-services":
+        uri = "http://169.254.169.254/latest/meta-data/local-ipv4"
+        result = _cloud_metadata_request(uri)
+        if result:
+            _ip_for_cloud = result
+            return [_ip_for_cloud]
+    return []
+
+
+def debug_timestamp():
+    return datetime.datetime.now().strftime('%Y/%m/%d %H:%M:%S')
+
+
 # vim:ts=4:sw=4:et:
