@@ -17,13 +17,13 @@ from .msg import common_err, common_error, common_warn, common_debug, cib_parse_
 from . import userdir
 from .utils import add_sudo, str2file, str2tmp, get_boolean
 from .utils import get_stdout, stdout2list, crm_msec, crm_time_cmp
-from .utils import olist, get_cib_in_use, get_tempdir
+from .utils import olist, get_cib_in_use, get_tempdir, to_ascii
 
 
 def xmlparse(f):
     try:
         cib_elem = etree.parse(f).getroot()
-    except Exception, msg:
+    except Exception as msg:
         common_err("cannot parse xml: %s" % msg)
         return None
     return cib_elem
@@ -32,13 +32,13 @@ def xmlparse(f):
 def file2cib_elem(s):
     try:
         f = open(s, 'r')
-    except IOError, msg:
+    except IOError as msg:
         common_err(msg)
         return None
     cib_elem = xmlparse(f)
     f.close()
     if options.regression_tests and cib_elem is None:
-        print "Failed to read CIB from file: %s" % (s)
+        print("Failed to read CIB from file: %s" % (s))
     return cib_elem
 
 
@@ -51,14 +51,14 @@ def compressed_file_to_cib(s):
             f = gzip.open(s)
         else:
             f = open(s)
-    except IOError, msg:
+    except IOError as msg:
         common_err(msg)
         return None
     cib_elem = xmlparse(f)
     if options.regression_tests and cib_elem is None:
-        print "Failed to read CIB from file %s" % (s)
+        print("Failed to read CIB from file %s" % (s))
         f.seek(0)
-        print f.read()
+        print(f.read())
     f.close()
     return cib_elem
 
@@ -69,13 +69,13 @@ cib_dump = "cibadmin -Ql"
 def sudocall(cmd):
     cmd = add_sudo(cmd)
     if options.regression_tests:
-        print ".EXT", cmd
+        print(".EXT", cmd)
     p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
     try:
         outp, errp = p.communicate()
         p.wait()
-        return p.returncode, outp, errp
-    except IOError, msg:
+        return p.returncode, to_ascii(outp), to_ascii(errp)
+    except IOError as msg:
         common_err("running %s: %s" % (cmd, msg))
         return None, None, None
 
@@ -87,12 +87,20 @@ def cibdump2file(fname):
     return None
 
 
-def cibdump2tmp():
+def cibdump2tmp(filterfn=None):
     try:
         _, outp, _ = sudocall(cib_dump)
+        if filterfn is not None:
+            try:
+                cib_elem = etree.fromstring(outp)
+            except etree.ParseError as msg:
+                common_err(msg)
+                return None
+            filterfn(cib_elem)
+            outp = etree.tostring(cib_elem, pretty_print=True)
         if outp is not None:
             return str2tmp(outp)
-    except IOError, msg:
+    except IOError as msg:
         common_err(msg)
     return None
 
@@ -104,7 +112,7 @@ def cibtext2elem(cibtext):
     """
     try:
         return etree.fromstring(cibtext)
-    except Exception, err:
+    except Exception as err:
         cib_parse_err(err, cibtext)
         return None
 
@@ -334,12 +342,15 @@ def mk_rsc_type(n):
     return ''.join((s1, s2, ra_type))
 
 
-def listnodes():
+def listnodes(include_remote_nodes=True):
     cib = cibdump2elem()
     if cib is None:
         return []
     local_nodes = cib.xpath('/cib/configuration/nodes/node/@uname')
-    remote_nodes = cib.xpath('/cib/status/node_state[@remote_node="true"]/@uname')
+    if include_remote_nodes:
+        remote_nodes = cib.xpath('/cib/status/node_state[@remote_node="true"]/@uname')
+    else:
+        remote_nodes = []
     return list(set([n for n in local_nodes + remote_nodes if n]))
 
 
@@ -400,17 +411,17 @@ def shadowfile(name):
 def pe2shadow(pe_file, name):
     '''Copy a PE file (or any CIB file) to a shadow.'''
     try:
-        s = open(pe_file).read()
-    except IOError, msg:
+        bits = open(pe_file, 'rb').read()
+    except IOError as msg:
         common_err("open: %s" % msg)
         return False
     # decompresed if it ends with .bz2
     if pe_file.endswith(".bz2"):
-        s = bz2.decompress(s)
+        bits = bz2.decompress(bits)
     # copy input to the shadow
     try:
-        open(shadowfile(name), "w").write(s)
-    except IOError, msg:
+        open(shadowfile(name), "wb").write(bits)
+    except IOError as msg:
         common_err("open: %s" % msg)
         return False
     return True
@@ -452,7 +463,7 @@ def drop_attr_defaults(node, ts=0):
 
 def nameandid(e, level):
     if e.tag:
-        print level*' ', e.tag, e.get("id"), e.get("name")
+        print(level*' ', e.tag, e.get("id"), e.get("name"))
 
 
 def xmltraverse(e, fun, ts=0):
@@ -580,7 +591,7 @@ def is_defaults(node):
 
 
 def rsc_constraint(rsc_id, con_elem):
-    for attr in con_elem.keys():
+    for attr in list(con_elem.keys()):
         if attr in constants.constraint_rsc_refs \
                 and rsc_id == con_elem.get(attr):
             return True
@@ -639,7 +650,7 @@ def printid(e_list):
     for e in e_list:
         ident = e.get("id")
         if ident:
-            print "element id:", ident
+            print("element id:", ident)
 
 
 def remove_dflt_attrs(e_list):
@@ -649,7 +660,7 @@ def remove_dflt_attrs(e_list):
     for e in e_list:
         try:
             d = constants.attr_defaults[e.tag]
-            for a in d.keys():
+            for a in list(d.keys()):
                 if e.get(a) == d[a]:
                     del e.attrib[a]
         except:
@@ -664,6 +675,7 @@ def remove_text(e_list):
 
 
 def sanitize_cib(doc):
+
     xml_processnodes(doc, is_status_node, rmnodes)
     # xml_processnodes(doc, true, printid)
     # xml_processnodes(doc, is_emptynvpairs, rmnodes)
@@ -675,6 +687,17 @@ def sanitize_cib(doc):
     xml_processnodes(doc, true, remove_text)
     xmltraverse(doc, drop_attr_defaults)
 
+def sanitize_cib_for_patching(doc):
+    """
+    Custom version of sanitize_cib which
+    doesn't sort container children, to use
+    for processing the original CIB when
+    generating a patch to apply using crm_diff.
+    """
+    xml_processnodes(doc, is_status_node, rmnodes)
+    xml_processnodes(doc, is_entity, rmnodes)
+    xml_processnodes(doc, true, remove_dflt_attrs)
+    xml_processnodes(doc, true, remove_text)
 
 def is_simpleconstraint(node):
     return len(node.xpath("resource_set/resource_ref")) == 0
@@ -800,7 +823,7 @@ def get_op_timeout(rsc_node, op, default_timeout):
 def op2list(node):
     pl = []
     action = ""
-    for name in node.keys():
+    for name in list(node.keys()):
         if name == "name":
             action = node.get(name)
         elif name != "id":  # skip the id
@@ -820,7 +843,7 @@ def get_rsc_operations(rsc_node):
 def make_sort_map(*order):
     m = {}
     for i, o in enumerate(order):
-        if isinstance(o, basestring):
+        if isinstance(o, str):
             m[o] = i
         else:
             for k in o:
@@ -829,7 +852,7 @@ def make_sort_map(*order):
 
 
 _sort_xml_order = make_sort_map('node',
-                                'template', 'primitive', 'group', 'master', 'clone', 'op',
+                                'template', 'primitive', 'bundle', 'group', 'master', 'clone', 'op',
                                 'tag',
                                 ['rsc_location', 'rsc_colocation', 'rsc_order'],
                                 ['rsc_ticket', 'fencing-topology'],
@@ -838,7 +861,7 @@ _sort_xml_order = make_sort_map('node',
                                 'alert')
 
 _sort_cli_order = make_sort_map('node',
-                                'rsc_template', 'primitive', 'group',
+                                'rsc_template', 'primitive', 'bundle', 'group',
                                 ['ms', 'master'], 'clone', 'op',
                                 'tag',
                                 ['location', 'colocation', 'collocation', 'order'],
@@ -858,7 +881,7 @@ def processing_sort(nl):
     TODO: if sort_elements is disabled, only sort to resolve inter-dependencies.
     '''
     def sort_elements(k):
-        return _sort_xml_order.get(k.tag, _SORT_LAST), k.get('id')
+        return _sort_xml_order.get(k.tag, _SORT_LAST)
 
     def sort_type(k):
         return _sort_xml_order.get(k.tag, _SORT_LAST)
@@ -917,7 +940,7 @@ def rename_id(node, old_id, new_id):
 
 def rename_rscref_simple(c_obj, old_id, new_id):
     c_modified = False
-    for attr in c_obj.node.keys():
+    for attr in list(c_obj.node.keys()):
         if attr in constants.constraint_rsc_refs and \
                 c_obj.node.get(attr) == old_id:
             c_obj.node.set(attr, new_id)
@@ -928,7 +951,7 @@ def rename_rscref_simple(c_obj, old_id, new_id):
 
 def delete_rscref_simple(c_obj, rsc_id):
     c_modified = False
-    for attr in c_obj.node.keys():
+    for attr in list(c_obj.node.keys()):
         if attr in constants.constraint_rsc_refs and \
                 c_obj.node.get(attr) == rsc_id:
             del c_obj.node.attrib[attr]
@@ -1058,7 +1081,7 @@ def silly_constraint(c_node, rsc_id):
         return cnt <= 1
     cnt = 0  # total count of referenced resources have to be at least two
     rsc_cnt = 0
-    for attr in c_node.keys():
+    for attr in list(c_node.keys()):
         if attr in constants.constraint_rsc_refs:
             cnt += 1
             if c_node.get(attr) == rsc_id:
@@ -1188,14 +1211,14 @@ def set_attr(e, attr, value):
     '''
     Set an attribute in the attribute set.
     '''
-    nvpair = get_attr_in_set(e, attr)
-    if nvpair is None:
+    nvp = get_attr_in_set(e, attr)
+    if nvp is None:
         from . import idmgmt
-        nvpair = etree.SubElement(e, "nvpair", id="", name=attr, value=value)
-        nvpair.set("id", idmgmt.new(nvpair, e.get("id")))
+        nvp = etree.SubElement(e, "nvpair", id="", name=attr, value=value)
+        nvp.set("id", idmgmt.new(nvp, e.get("id")))
     else:
-        nvpair.set("name", attr)
-        nvpair.set("value", value)
+        nvp.set("name", attr)
+        nvp.set("value", value)
 
 
 def get_set_nodes(e, setname, create=False):
@@ -1226,12 +1249,12 @@ def xml_equals_unordered(a, b):
         return False
 
     def tagflat(x):
-        return isinstance(x.tag, basestring) and x.tag or x.text
+        return isinstance(x.tag, str) and x.tag or x.text
 
     def sortby(v):
         if v.tag == 'primitive':
             return v.tag
-        return tagflat(v) + ''.join(sorted(v.attrib.keys() + v.attrib.values()))
+        return tagflat(v) + ''.join(sorted(list(v.attrib.keys()) + list(v.attrib.values())))
 
     def safe_strip(text):
         return text is not None and text.strip() or ''
@@ -1254,7 +1277,7 @@ def xml_equals_unordered(a, b):
     if a.tag == 'resource_set':
         return all(xml_equals_unordered(a, b) for a, b in zip(a, b))
     else:
-        sorted_children = zip(sorted(a, key=sortby), sorted(b, key=sortby))
+        sorted_children = list(zip(sorted(a, key=sortby), sorted(b, key=sortby)))
         return all(xml_equals_unordered(a, b) for a, b in sorted_children)
 
 
@@ -1263,19 +1286,28 @@ def xml_equals(n, m, show=False):
     if not rc and show and config.core.debug:
         # somewhat strange, but that's how this works
         from doctest import Example
-        example = Example("etree.tostring(n)", etree.tostring(n))
-        got = etree.tostring(m)
-        print _checker.output_difference(example, got, 0)
+        example = Example("etree.tostring(n)", xml_tostring(n))
+        got = xml_tostring(m)
+        print(_checker.output_difference(example, got, 0))
     return rc
 
 
+def xml_tostring(*args, **kwargs):
+    """
+    Python 2/3 conversion utility:
+    etree.tostring returns a bytestring, but
+    we need actual Python strings.
+    """
+    return etree.tostring(*args, **kwargs).decode('utf-8')
+
+
 def merge_attributes(dnode, snode, tag):
     rc = False
     add_children = []
     for sc in snode.iterchildren(tag):
         dc = lookup_node(sc, dnode, ignore_id=True)
         if dc is not None:
-            for a, v in sc.items():
+            for a, v in list(sc.items()):
                 if a == "id":
                     continue
                 if v != dc.get(a):
@@ -1324,7 +1356,7 @@ def merge_tmpl_into_prim(prim_node, tmpl_node):
     merge_nodes(dnode, tmpl_node)
     merge_nodes(dnode, prim_node)
     # the resulting node should inherit all primitives attributes
-    for a, v in prim_node.items():
+    for a, v in list(prim_node.items()):
         dnode.set(a, v)
     # but class/provider/type are coming from the template
     # savannah#41410: stonith resources do not have the provider
